"""A collection of Callable classes to be used by Processors and TransitionPredicates.
"""
from gabrieltool.statemachine.callable_zoo.base import record_kwargs, CallableBase, Null  # noqa: F401
from gabrieltool.statemachine.callable_zoo import processor_zoo  # noqa: F401
from gabrieltool.statemachine.callable_zoo import predicate_zoo  # noqa: F401



"""Utilities for using Tensorflow models.
"""
import numpy as np
import tensorflow as tf
import grpc
from tensorflow_serving.apis import predict_pb2, prediction_service_pb2_grpc


class TFServingPredictor(object):
    """An agent that makes request to a TF serving server to get object detection results.

    This agent communicates with the TF serving server (often a container at
    localhost) through gRPC.
    """

    def __init__(self, host, port):
        """Constructor.

        Args:
            host (string): TF serving server hostname or IP address.
            port (int): TF serving server port number.
        """
        options = [
            ('grpc.max_message_length', 500 * 1024 * 1024),
            ('grpc.max_send_message_length', 500 * 1024 * 1024),
            ('grpc.max_receive_message_length', 500 * 1024 * 1024),
        ]
        self.channel = grpc.insecure_channel('{}:{}'.format(host, int(port)), options=options)
        self.stub = prediction_service_pb2_grpc.PredictionServiceStub(self.channel)

    def infer_one(self, model_name, rgb_image, conf_threshold=0.5):
        """Infer one image by sending a request to TF serving server.

        Args:
            model_name (string): Name of the Model
            rgb_image (numpy array): Image in RGB format
            conf_threshold (float, optional): Cut-off threshold for detection. Defaults to 0.5.

        Returns:
            Dictionary: keys are class ids, values are list of [x1, y1, x2,
            y2, confidence, label_idx]. e.g {'cat': [[0, 0, 100, 100, 0.6, 'cat']],
            1: [[0, 0, 100, 100, 0.7, 1]]}
        """
        parsed_results = self._infer(model_name, np.stack([rgb_image.astype(dtype=np.uint8)], axis=0))
        results = {}
        # parsed_results has
        # num_detections: number of detections
        # detection_scores: 2d array of confidence, [image_idx, bbx_idx]
        # detection_classes: 2d array, [image_idx, bbx_idx]
        # detection_boxes: 3d array, [image_idx, bbx_idx, (ymin,xmin,ymax,xmax)]
        # for each image, should only be 1 image here.
        # num_detections = parsed_results['num_detections']
        detection_scores = parsed_results['detection_scores']
        detection_classes = parsed_results['detection_classes']
        detection_boxes = parsed_results['detection_boxes']
        h, w, _ = rgb_image.shape
        image_idx = 0
        for detection_idx in range(len(detection_classes[image_idx])):
            label = str(detection_classes[image_idx][detection_idx])
            confidence = detection_scores[image_idx][detection_idx]
            if confidence < conf_threshold:
                continue
            norm_bbox = detection_boxes[image_idx][detection_idx]
            bbox = [int(norm_bbox[1]*w), int(norm_bbox[0]*h),
                    int(norm_bbox[3]*w), int(norm_bbox[2]*h)]
            objs = results.setdefault(label, [])
            objs.append([*bbox, confidence, label])
        return results

    def _infer(self, model_name, images):
        # Create prediction request object
        request = predict_pb2.PredictRequest()
        # Specify model name (must be the same as when the TensorFlow serving serving was started)
        request.model_spec.name = model_name
        # Initalize prediction
        request.inputs['inputs'].CopyFrom(
            tf.make_tensor_proto(images))
        # Call the prediction server
        result = self.stub.Predict(request, 10.0)  # 10 secs timeout
        # convert tensorProto to numpy array
        parsed_results = {}
        for k, v in result.outputs.items():
            parsed_results[k] = tf.make_ndarray(v)
        # fix output result types
        if 'detection_classes' in parsed_results:
            parsed_results['detection_classes'] = parsed_results['detection_classes'].astype(np.int64)
        return parsed_results



"""A collection of Callable classes to be used by Processors (in FSM states).
"""
from .base import DummyCallable, FasterRCNNOpenCVCallable  # noqa: F401
from .containerized import FasterRCNNContainerCallable  # noqa: F401
from .containerized import TFServingContainerCallable  # noqa: F401



# -*- coding: utf-8 -*-
"""Components to Create a Finite State Machine.

See FSM's wikipedia page for its basics:
https://en.wikipedia.org/wiki/Finite-state_machine.

This modules provides components to create, edit, serialize, and deserialize a
finite state machine. Below is a list of key concepts.

* State: FSM states represents the status of a cognitive assistant. States
    have Processors, which are executed to analyze the input data when the
    application is in the state.
* Transitions: Transitions define the conditions (TransitionPredicate) for
    state change and actions (Instruction) to take when changing states.
* Finite State Machine (StateMachine): An FSM is a set of states and
    transitions. Helper functions are provided in the StateMachine class for
    serialization, deserialization and traversal.
"""


import json

from gabrieltool.statemachine import (predicate_zoo, processor_zoo,
                                      callable_zoo,
                                      wca_state_machine_pb2)


class _FSMObjBase:
    """Base class for all FSM component classes.

    This base class serves as an adapter to help serialize/deserialize
    components into the protobuf message formats defined in
    proto/wca-state-machine.proto. Variables defined in the proto are exposed as
    instance variables.
    """
    _obj_cnt = 0

    def __init__(self, name=None):
        """Constructor

        A protobuf message is created to store all the variables for
        serialization.

        Args:
            name (string, optional): name of the component. Defaults to None.
        """

        super().__init__()
        _FSMObjBase._obj_cnt += 1
        protobuf_message = getattr(wca_state_machine_pb2,
                                   self.__class__.__name__)()
        self._pb = protobuf_message
        self._expose_serializer_attr('name', 'rw')
        self.name = self._get_default_name()
        if name is not None:
            self.name = name

    def _get_default_name(self):
        return '{}_{}'.format(self.__class__.__name__,
                              _FSMObjBase._obj_cnt)

    def __repr__(self):
        default = super(_FSMObjBase, self).__repr__()
        return '<{} ({})>'.format(self.name, default)

    def _expose_serializer_attr(self, name, mode):
        """Helper method to provide easy access to read and write the protobuf message as instance variables."""
        if mode == 'r':
            setattr(_FSMObjBase, name, property(
                lambda self: getattr(self._pb, name)))
        elif mode == 'rw':
            setattr(_FSMObjBase,
                    name,
                    property(lambda self: getattr(self._pb, name),
                             lambda self, value: setattr(
                                 self._pb, name, value)
                             ))
        else:
            raise ValueError(
                'Unsupported mode {}. Valid modes are "r" or "rw"'.format(mode))

    def from_desc(self, desc):
        """Construct an object from its serialized description."""
        self._pb = desc

    def to_desc(self):
        """Returned the serialized description of this object as a protobuf message."""
        return self._pb


class _FSMCallable(_FSMObjBase):
    """A callable FSM component.

    This serves as a base class for FSM components that needs to implement a
    callable interface (e.g. Processor, TransitionPredicate).
    """

    def __init__(self, name=None, callable_obj=None, zoo=None):
        super().__init__(name)
        self._callable_zoo = zoo
        self.callable_obj = callable_obj if callable_obj is not None else callable_zoo.Null()

    @property
    def callable_obj(self):
        """The callable object to invoke when this object is called."""
        return self._callable_obj

    @callable_obj.setter
    def callable_obj(self, obj):
        if not isinstance(obj, callable_zoo.CallableBase):
            raise TypeError(
                'Invalid type ({}). '
                '_FSMCallable\'s callable_obj requires '
                'a callable_zoo.CallableBase object.'.format(type(obj)))
        self._callable_obj = obj

    def prepare(self):
        """Invoke prepare() method of the callable_obj if it has any.

        This prepare method is called when the FSM runner starts executing to
        give callables an opportunity to initialize themselves.
        """
        prepare_func = getattr(self._callable_obj, "prepare", None)
        if callable(prepare_func):
            prepare_func()

    def __call__(self, current_input):
        return self._callable_obj(current_input)

    def from_desc(self, data):
        super().from_desc(data)
        callable_class = getattr(self._callable_zoo, self._pb.callable_name)
        initializer_args = json.loads(self._pb.callable_args)
        self._callable_obj = callable_class.from_json(initializer_args)

    def to_desc(self):
        self._pb.callable_name = self._callable_obj.__class__.__name__
        self._pb.callable_args = json.dumps(self._callable_obj.kwargs)
        return super().to_desc()


class Processor(_FSMCallable):
    """Processor specifies how to process input (e.g. an image) in a state.
    """

    def __init__(self, name=None, callable_obj=None):
        """Construct a processor.

        Args:
            name (string, optional): Name of the processor. Defaults to None.
            callable_obj (subclass of callable_zoo.CallableBase, optional): An
                object whose type is a subclass of callable_zoo.CallableBase.
                This object is called/executed when there is a new input (e.g. an image).
                This callable_obj should expect exact one positional argument. Defaults to None.
        """
        super().__init__(name=name, callable_obj=callable_obj, zoo=processor_zoo)


class TransitionPredicate(_FSMCallable):
    """Condition for state transition.

    TransitionPredicate determines whether a state transition should taken.
    TransitionPredciate implements the callable interface so that its objects
    can be evaluated as a function. A state transition is taken when a
    TransitionPredicate evaluates to True.
    """

    def __init__(self, name=None, callable_obj=None):
        """Construct a transition predicate.

        Args:
            name (string, optional): Name of the TransitionPredicate. Defaults to None.
            callable_obj (subclass of callable_zoo.CallableBase, optional): An
                object whose type is a subclass of callable_zoo.CallableBase. This
                object is called/executed when this transition predicate is called
                to determine whether the current transition should be taken.
                This callable_obj should expect exact one positional argument of type dictionary,
                which contains the output of current state's processors. Defaults to None.
        """
        super().__init__(name=name, callable_obj=callable_obj, zoo=predicate_zoo)


class Instruction(_FSMObjBase):
    """Instruction to return when a transition is taken."""

    def __init__(self, name=None, audio=None, image=None, video=None):
        """Instructions can have audio, image, or video.

        Args:
            name (string, optional): Name of the instruction. Defaults to None.
            audio (string, optional): Verbal instruction in text. Defaults to None.
            image (bytes, optional): Encoded Image in bytes. Defaults to None.
            video (url string, optional): Video Url in string. Defaults to None.
        """
        super(Instruction, self).__init__(name=name)
        expose_attrs = [('audio', 'rw'), ('image', 'rw'), ('video', 'rw')]
        for (attr, mode) in expose_attrs:
            self._expose_serializer_attr(attr, mode)
        if audio is not None:
            self.audio = audio
        if image is not None:
            self.image = image
        if video is not None:
            self.video = video


class Transition(_FSMObjBase):
    """Links among FSM states that defines state changes and results to return when changing states.

    A Transition has the following components:
      * transition predicates: The conditions that need to be satisfied to take this transition.
      * next_state: The next FSM state to visit after taking the transition.
      * instructions: Instructions returned to users when this
        transition is taken.
    """

    def __init__(self, name=None, predicates=None, instruction=None, next_state=None):
        """Construct a Transition

        Args:
            name (string, optional): Name of the transition. Defaults to None.
            predicates (a list of TransitionPredicate, optional): A list of
                condition to satisfy. They are daisy-chained (AND) together when
                evaluating whether this transition takes place. Defaults to None.
            instruction (Instruction, optional): Instruction to give. Defaults to None.
            next_state (State, optional): Next state to move to. Defaults to None.
        """
        super(Transition, self).__init__(name)
        self.predicates = predicates if predicates is not None else []
        self.instruction = instruction if instruction is not None else Instruction()
        self.next_state = next_state

    @property
    def predicates(self):
        """The list of TransitionPredicates."""
        return self._predicates

    @predicates.setter
    def predicates(self, val):
        if type(val) != list:
            raise TypeError("Predicates needs to be type list.")
        self._predicates = val

    def __call__(self, app_state):
        """Given the current app_state, check if a transition should be taken.

        The condition of taking a transition is that all the trigger predicates
        are satisfied. When all of them are satisfied, return True, next state,
        and instruction. Otherwise, return False, None, None.
        """
        for predicate in self._predicates:
            if not predicate(app_state):
                return None
        return self

    def to_desc(self):
        for pred in self._predicates:
            self._pb.predicates.extend([pred.to_desc()])
        if self.instruction is not None:
            self._pb.instruction.CopyFrom(self.instruction.to_desc())
        if self.next_state:
            self._pb.next_state = self.next_state.name
        return super(Transition, self).to_desc()

    def from_desc(self):
        """Do not call this method directly.

        Transition itself does not know enough information to build from its
        description. next_state variable depends on a FSM. Use StateMachine Helper Class instead.

        Raises:
            NotImplementedError: Always
        """
        raise NotImplementedError("Transition itself does not know enough information "
                                  "to build from its description. "
                                  "next_state variable depends on a FSM. "
                                  "Use StateMachine Helper Class instead.")


class State(_FSMObjBase):
    """A FSM state represents the status of the system.

    A state can have many processors and transitions.
    """

    def __init__(self, name=None, processors=None, transitions=None):
        """Construct a FSM state.

        Args:
            name (string, required): Name of the State. Each state in a FSM needs to
                have an unique name.
            processors (list of Processor, optional): Processor to run on an
                input in this state. Each processor will be called with exactly one
                positional argument (input), and should return a dictionary that
                contains the extracted information. The returned dictionaries from
                multiple processors will be unioned together to serve as inputs to
                transition predicates. Defaults to None.
            transitions (list of Transition, optional): Possible Transitions
                from this state. Transitions are evaluated one by one in the order
                of this list. The first transition that satisfies will be taken.
                Defaults to None.
        """
        super(State, self).__init__(name)
        self.processors = processors if processors is not None else []
        self.transitions = transitions if transitions is not None else []

    @property
    def processors(self):
        """The list of processors to be executed in this state."""
        return self._processors

    @processors.setter
    def processors(self, val):
        if type(val) != list:
            raise TypeError("Predicates needs to be type list.")
        self._processors = val

    @property
    def transitions(self):
        """The list of possible transitions to take in this state."""
        return self._transitions

    @transitions.setter
    def transitions(self, val):
        if type(val) != list:
            raise TypeError("Predicates needs to be type list.")
        self._transitions = val

    def _run_processors(self, img):
        app_state = {'raw': img}
        for obj_processor in self.processors:
            app_state.update(obj_processor(img))
        return app_state

    def _get_one_satisfied_transition(self, app_state):
        for transition in self.transitions:
            if transition(app_state) is not None:
                return transition
        return None

    def prepare(self):
        """Prepare a state (e.g. initialize all processors and transition predicates.)

        This method is called when the FSM runner first starts to
        give callables an opportunity to initialize themselves.
        """
        for obj_processor in self.processors:
            obj_processor.prepare()

    def __call__(self, img):
        app_state = self._run_processors(img)
        transition = self._get_one_satisfied_transition(app_state)
        if transition is None:
            return self, Instruction()
        else:
            return transition.next_state, transition.instruction

    def to_desc(self):
        for proc in self.processors:
            self._pb.processors.extend([proc.to_desc()])
        for tran in self.transitions:
            self._pb.transitions.extend([tran.to_desc()])
        return super(State, self).to_desc()

    def from_desc(self):
        """Do not call this method directly.

        State itself does not know enough information to build from its
        description. The next_state in state's transitions depends on a FSM. Use StateMachine Helper Class instead.

        Raises:
            NotImplementedError: Always
        """
        raise NotImplementedError("State itself does not know enough information "
                                  "to build from its description. "
                                  "next_state variable depends on a FSM. "
                                  "Use StateMachine Helper Class instead.")


class StateMachine(object):
    """Helper class to serialize, deserialize, and traverse a state machine.
    """

    @classmethod
    def _load_generic_from_desc(cls, GenericType, desc):
        item = GenericType()
        item.from_desc(desc)
        return item

    @classmethod
    def _load_transition(cls, desc, state_lut):
        tran = Transition(name=desc.name)
        tran.instruction = cls._load_generic_from_desc(Instruction,
                                                       desc.instruction)
        preds = [cls._load_generic_from_desc(
            TransitionPredicate, pred_desc)
            for pred_desc in desc.predicates]
        tran._predicates = preds
        tran.next_state = state_lut[desc.next_state]
        return tran

    @classmethod
    def _load_state(cls, desc, state_lut):
        state = state_lut[desc.name]
        state.processors = [cls._load_generic_from_desc(
            Processor, proc_desc) for proc_desc in desc.processors]
        state.transitions = [cls._load_transition(
            tran_desc, state_lut) for tran_desc in desc.transitions]
        return state

    @classmethod
    def from_bytes(cls, data):
        """Load a State Machine from bytes.

        Args:
            data (bytes): Serialized FSM in bytes. Format is specified in
            wca_state_machine.proto.

        Raises:
            ValueError: raised when there are duplicate state names.

        Returns:
            State: The start state of the FSM.
        """
        pb_fsm = wca_state_machine_pb2.StateMachine()
        pb_fsm.ParseFromString(data)
        state_lut = {}
        # 1st pass get all states
        for state_desc in pb_fsm.states:
            state = State(name=state_desc.name)
            if state.name in state_lut:
                raise ValueError(
                    "Duplicate State Name: {}. Invalid State Machine Data.".format(
                        state.name))
            state_lut[state.name] = state
        # 2nd pass to load all state details
        for state_desc in pb_fsm.states:
            cls._load_state(state_desc, state_lut)
        return state_lut[pb_fsm.start_state]

    @classmethod
    def bfs(cls, start_state):
        """Generator for a breadth-first traversal on the FSM.

        This method can be used to enumerate states in an FSM.

        Args:
            start_state (State): The start state of the traversal.

        Yields:
            State: The current state of the traversal.
        """
        visited = set([start_state])
        work_queue = [start_state]
        while work_queue:
            state = work_queue.pop(0)
            for tran in state.transitions:
                if tran.next_state is not None and tran.next_state not in visited:
                    work_queue.append(tran.next_state)
                    visited.add(tran.next_state)
            yield state

    @classmethod
    def to_bytes(cls, name, start_state):
        """Serialize a FSM to bytes.

        States in the FSM are discovered using a breadth-first search (see the
        bfs method in this class).

        Args:
            name (string): The name of the FSM.
            start_state (State): The start state of the FSM.

        Raises:
            ValueError: raised when there are duplicate state names.

        Returns:
            bytes: Serialized FSM in bytes. Format is defined in wca_state_machine.proto.
        """
        # TODO(junjuew) optimize/dedup assets/kwargs
        visited = {}
        for state in cls.bfs(start_state):
            if state.name in visited and visited[state.name] is not state:
                raise ValueError("Found duplicate state name {}. "
                                 "Cannot serialize a FSM with duplicate state name".format(state.name))
            visited[state.name] = state

        # serialize
        pb_fsm = wca_state_machine_pb2.StateMachine(
            name=name, start_state=start_state.name)
        for (state_name, state) in list(visited.items()):
            pb_fsm.states.extend([state.to_desc()])
        return pb_fsm.SerializeToString()



# -*- coding: utf-8 -*-
"""Basic callable classes for Processor.
"""
import copy

import cv2
import numpy as np
from logzero import logger

from gabrieltool.statemachine.callable_zoo import record_kwargs
from gabrieltool.statemachine.callable_zoo import CallableBase


def visualize_detections(img, results):
    """Visualize object detection outputs.

    This is a helper function for debugging processor callables.
    The results should follow Gabrieltool's convention, which is

    Arguments:
        img {OpenCV Image}
        results {Dictionary} -- a dictionary of class_idx -> [[x1, y1, x2, y2, confidence, cls_idx],...]

    Returns:
        OpenCV Image -- Image with detected objects annotated
    """
    img_detections = img.copy()
    for _, dets in results.items():
        for i in range(len(dets)):
            cls_name = str(dets[i][-1])
            bbox = dets[i][:4]
            score = dets[i][-2]
            text = "%s : %f" % (cls_name, score)
            cv2.rectangle(img_detections, (int(bbox[0]), int(bbox[1])), (int(bbox[2]), int(bbox[3])), (0, 0, 255), 8)
            cv2.putText(img_detections, text, (int(bbox[0]), int(bbox[1])), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 0, 0), 2)
    return img_detections


class DummyCallable(CallableBase):
    """A Dummy Callable class for testing and examples.
    """

    @record_kwargs
    def __init__(self, dummy_input='dummy_input_value'):
        super(DummyCallable, self).__init__()

    def __call__(self, image, debug=False):
        return {'dummy_key': 'dummy_value'}


class FasterRCNNOpenCVCallable(CallableBase):
    """A callable class that executes a FasterRCNN object detection model using OpenCV.
    """

    @record_kwargs
    def __init__(self, proto_path, model_path, labels=None, conf_threshold=0.8):
        """Constructor.

        Args:
            proto_path (string): Path to the caffe proto files that defines the DNN.
            model_path (string): Path to the model weights file.
            labels (list of string, optional): List of labels. Defaults to None.
            conf_threshold (float, optional): Confidence threshold for a detection. Defaults to 0.8.
        """
        # For default parameter settings,
        # see:
        # https://github.com/rbgirshick/fast-rcnn/blob/b612190f279da3c11dd8b1396dd5e72779f8e463/lib/fast_rcnn/config.py
        super(FasterRCNNOpenCVCallable, self).__init__()
        self._scale = 600
        self._max_size = 1000
        # Pixel mean values (BGR order) as a (1, 1, 3) array
        # We use the same pixel mean for all networks even though it's not exactly what
        # they were trained with
        self._pixel_means = [102.9801, 115.9465, 122.7717]
        self._nms_threshold = 0.3
        self._labels = labels
        self._net = cv2.dnn.readNetFromCaffe(proto_path, model_path)
        self._conf_threshold = conf_threshold
        logger.debug(
            'Created a FasterRCNNOpenCVProcessor:\nDNN proto definition is at {}\n'
            'model weight is at {}\nlabels are {}\nconf_threshold is {}'.format(
                proto_path, model_path, self._labels, self._conf_threshold))

    @classmethod
    def from_json(cls, json_obj):
        """Create an object from a JSON object.

        Args:
            json_obj (json): JSON object that has all the serialized constructor
                arguments.

        Raises:
            ValueError: when constructor arguments' type don't match.

        Returns:
            FasterRCNNOpenCVCallable: The deserialized FasterRCNNOpenCVCallable object.
        """
        try:
            kwargs = copy.copy(json_obj)
            kwargs['labels'] = json_obj['labels']
            kwargs['_conf_threshold'] = float(json_obj['conf_threshold'])
        except ValueError as e:
            raise ValueError(
                'Failed to convert json object to {} instance. '
                'The input json object is {}. ({})'.format(cls.__name__,
                                                           json_obj, e))
        return cls(**json_obj)

    def _getOutputsNames(self, net):
        layersNames = net.getLayerNames()
        return [layersNames[i[0] - 1] for i in net.getUnconnectedOutLayers()]

    def __call__(self, image):
        height, width = image.shape[:2]

        # resize image to correct size
        im_size_min = np.min(image.shape[0:2])
        im_size_max = np.max(image.shape[0:2])
        im_scale = float(self._scale) / float(im_size_min)
        # Prevent the biggest axis from being more than MAX_SIZE
        if np.round(im_scale * im_size_max) > self._max_size:
            im_scale = float(self._max_size) / float(im_size_max)
        im = cv2.resize(image, None, None, fx=im_scale, fy=im_scale,
                        interpolation=cv2.INTER_LINEAR)
        # create input data
        blob = cv2.dnn.blobFromImage(im, 1, (width, height), self._pixel_means,
                                     swapRB=False, crop=False)
        imInfo = np.array([height, width, im_scale], dtype=np.float32)
        self._net.setInput(blob, 'data')
        self._net.setInput(imInfo, 'im_info')

        # infer
        outs = self._net.forward(self._getOutputsNames(self._net))
        t, _ = self._net.getPerfProfile()
        logger.debug('Inference time: %.2f ms' % (t * 1000.0 / cv2.getTickFrequency()))

        # postprocess
        classIds = []
        confidences = []
        boxes = []
        for out in outs:
            for detection in out[0, 0]:
                confidence = detection[2]
                if confidence > self._conf_threshold:
                    left = int(detection[3])
                    top = int(detection[4])
                    right = int(detection[5])
                    bottom = int(detection[6])
                    width = right - left + 1
                    height = bottom - top + 1
                    classIds.append(int(detection[1]) - 1)  # Skip background label
                    confidences.append(float(confidence))
                    boxes.append([left, top, width, height])

        indices = cv2.dnn.NMSBoxes(boxes, confidences, self._conf_threshold, self._nms_threshold)
        results = {}
        for i in indices:
            i = i[0]
            box = boxes[i]
            left = box[0]
            top = box[1]
            width = box[2]
            height = box[3]
            classId = int(classIds[i])
            confidence = confidences[i]
            if self._labels[classId] not in results:
                results[self._labels[classId]] = []
            results[self._labels[classId]].append([left, top, left+width, top+height, confidence, classId])

        logger.debug('results: {}'.format(results))
        return results



"""Library to build application logic as a Finite State Machine.

_pb2 modules are generated python modules by Protobuf for serialization.
See the proto directory for the serialization formats.
"""

from gabrieltool.statemachine.callable_zoo import processor_zoo, predicate_zoo  # noqa: F401



"""Callable classes for Containerized Processors.

Currently we don't offer functionalities to clean up the containers after the
program finishes. Use the following commands to clean up the containers started
by this module.

    $ docker stop -t 0 $(docker ps -a -q --filter="name=GABRIELTOOL")

"""
import ast
import copy
import io
import os
import time

import cv2
import docker
import requests
from logzero import logger

from gabrieltool.statemachine.callable_zoo import record_kwargs
from gabrieltool.statemachine.callable_zoo import CallableBase
from gabrieltool.statemachine.callable_zoo.processor_zoo import tfutils

docker_client = docker.from_env()


class SingletonContainerManager():
    """Helper class to start, get, and remove a container identified by a name."""

    def __init__(self, container_name):
        self._container_name = container_name
        self._container = None

    @property
    def container_name(self):
        return self._container_name

    @property
    def container(self):
        if self._container is None:
            self._container = self._get_container_obj()
        return self._container

    def _add_image_tag_if_not_available(self, image_url):
        """default to latest tag if tag is not specified."""
        if ':' not in image_url:
            image_url = '{}:{}'.format(image_url, 'latest')
        return image_url

    def _make_image_available(self, image_url):
        logger.info(
            'Checking if the container image is available locally... (image: {})'.format(image_url))
        try:
            docker_client.images.get(image_url)
            logger.info(
                'Found local image (image: {})!'.format(image_url))
        except docker.errors.ImageNotFound:
            logger.info('{} not found locally. Trying to downloading it...'.format(image_url))
            logger.info('Downloading image: {}...'.format(image_url))
            repo = image_url.split(':')[0]
            tag = image_url.split(':')[-1]
            docker_client.images.pull(repo, tag=tag)
            logger.info('Download finished!')

    def start_container(self, image_url, command, **kwargs):
        """Start a container

        Args:
            image_url (string): Container Image URL.
            command (string): Container command.
            kwargs (dictionary): Extra arguments to pass to Docker client.

        Returns:
            Container: A container
        """
        image_url = self._add_image_tag_if_not_available(image_url)
        if self.container is None or self.container.status != 'running':
            self._make_image_available(image_url)
            # start container
            logger.info(
                'launching Docker container (name: {}, image: {}, command: {}, extra args: {})'.format(
                    self.container_name,
                    image_url,
                    command,
                    str(kwargs)
                ))
            try:
                container = docker_client.containers.run(
                    image_url,
                    command=command,
                    name=self.container_name,
                    auto_remove=True,
                    detach=True,
                    **kwargs
                )
                container.reload()
                # sleep here to give some time for container to start
                time.sleep(10)
                self._container = container
            except Exception as e:
                logger.error('Error starting container: {}'.format(e))
        return self._container

    def _get_container_obj(self):
        containers = docker_client.containers.list(filters={'name': self.container_name})
        if len(containers) > 0:
            return containers[0]
        return None

    def clean(self):
        """Remove the container if it exists."""
        container = self._get_container_obj()
        if container:
            container.remove(force=True)


class FasterRCNNContainerCallable(CallableBase):
    """A callable class to execute containerized FasterRCNN model in Caffe.

    Use this class if your object detector is generated by TPOD v1 and the
    container image is hosted by cmusatyalab's gitlab container registry.
    """
    # UNIQUE to each python process
    CONTAINER_NAME = 'GABRIELTOOL-FasterRCNNContainerCallable-{}'.format(os.getpid())

    @record_kwargs
    def __init__(self, container_image_url, conf_threshold=0.5):
        """Constructor.

        Args:
            container_image_url (string): URL to the container image.
            conf_threshold (float, optional): Cutoff threshold for detection. Defaults to 0.5.
        """
        # For default parameter settings,
        # see:
        # https://github.com/rbgirshick/fast-rcnn/blob/b612190f279da3c11dd8b1396dd5e72779f8e463/lib/fast_rcnn/config.py
        super(FasterRCNNContainerCallable, self).__init__()
        self.container_image_url = container_image_url
        self.conf_threshold = conf_threshold
        self.container_manager = SingletonContainerManager(self.CONTAINER_NAME)

        # start container
        # port number inside the container that is open
        self.container_port = '8000/tcp'
        ports = {self.container_port: None}   # map container 8000 to a random host port
        command = '/bin/bash run_server.sh',
        self.container_manager.start_container(self.container_image_url, command, ports=ports)

    @property
    def container_server_url(self):
        if self.container_manager.container is not None and self.container_manager.container.status == 'running':
            return 'http://localhost:{}/detect'.format(
                self.container_manager.container.ports[self.container_port][0]['HostPort'])
        return None

    @classmethod
    def from_json(cls, json_obj):
        """Deserialize."""
        try:
            kwargs = copy.copy(json_obj)
            kwargs['container_image_url'] = json_obj['container_image_url']
            kwargs['conf_threshold'] = float(json_obj['conf_threshold'])
        except ValueError as e:
            raise ValueError(
                'Failed to convert json object to {} instance. '
                'The input json object is {}. ({})'.format(cls.__name__,
                                                           json_obj, e))
        return cls(**kwargs)

    def __call__(self, image):
        fp = io.BytesIO()
        fp.write(cv2.imencode('.jpg', image)[1].tostring())
        fp.seek(0, 0)

        response = requests.post(self.container_server_url, data={
            'confidence': self.conf_threshold,
            'format': 'box'
        }, files={
            'picture': fp
        })
        detections = ast.literal_eval(response.text)
        result = {}
        for detection in detections:
            logger.info(detection)
            label = detection[0]
            bbox = detection[1]
            confidence = detection[2]
            if label not in result:
                result[label] = []
            result[label].append(
                [*bbox, confidence, label]
            )
        return result

    def clean(self):
        self.container_manager.clean()


class TFServingContainerCallable(CallableBase):
    """A callable class to execute frozen tensorflow models using TF serving container images.

    Use this class if your object detector is generated by OpenTPOD and you have
    downloaded the model. The TF serving container is started lazily when an
    FSM runner starts.
    """

    # UNIQUE to each python process
    CONTAINER_NAME = 'GABRIELTOOL-TFServingContainerCallable-{}'.format(os.getpid())
    # TF Serving image by default listens on 8500 for GRPC
    TFSERVING_GRPC_PORT = 8500
    SERVED_DIRS = {}

    @record_kwargs
    def __init__(self, model_name, serving_dir, conf_threshold=0.5):
        """Constructor.

        Args:
            model_name (string): Arbitrary Name of the Model. Each DNN should
                have a unique model name.
            serving_dir (string): Path to the TF saved_model. This should refers
                to the 'saved_model' directory of the downloaded OpenTPOD model.
            conf_threshold (float, optional): Cutoff threshold for detection. Defaults to 0.5.
        """
        super(TFServingContainerCallable, self).__init__()
        self.serving_dir = serving_dir
        self.model_name = model_name
        self.conf_threshold = conf_threshold
        TFServingContainerCallable.SERVED_DIRS[model_name] = os.path.abspath(serving_dir)
        self.container_manager = SingletonContainerManager(TFServingContainerCallable.CONTAINER_NAME)
        self.container_internal_port = '{}/tcp'.format(TFServingContainerCallable.TFSERVING_GRPC_PORT)
        self.predictor = None

    def prepare(self):
        """Launch the TF serving container. Do not call this method directly unless debugging.

        This function is called when an FSM runner starts. This enables
        gabrieltool to start only one TF serving container to serve many models.
        """
        # launch container
        self._start_container()

    def _start_container(self):
        """Launch TF serving container image to serve all entries in SERVED_DIRS."""
        ports = {self.container_internal_port: ('127.0.0.1', None)}
        container_image_url = 'tensorflow/serving:2.1.0'
        # geerate model config
        from tensorflow_serving.config import model_server_config_pb2
        from google.protobuf import text_format
        model_server_config = model_server_config_pb2.ModelServerConfig()
        for model_name, model_dir in TFServingContainerCallable.SERVED_DIRS.items():
            model_config = model_server_config.model_config_list.config.add()
            model_config.name = model_name
            model_config.base_path = '/models/{}'.format(model_name)
            model_config.model_platform = 'tensorflow'
        with open('models.config', 'w') as f:
            f.write(text_format.MessageToString(model_server_config))
        # mount volumes
        volumes = {
            os.path.abspath('models.config'): {'bind': '/models/models.config', 'mode': 'ro'}
        }
        for model_name, model_dir in TFServingContainerCallable.SERVED_DIRS.items():
            volumes[model_dir] = {'bind': '/models/{}'.format(model_name), 'mode': 'ro'}
        logger.debug('volumes: {}'.format(volumes))
        cmd = '--model_config_file=/models/models.config'
        self.container_manager.start_container(container_image_url, cmd, ports=ports, volumes=volumes)

    @property
    def container_external_port(self):
        """Port of the TF Serving container."""
        if self.container_manager.container is not None and self.container_manager.container.status == 'running':
            return self.container_manager.container.ports[self.container_internal_port][0]['HostPort']
        return None

    @classmethod
    def from_json(cls, json_obj):
        """Deserialize."""
        try:
            kwargs = copy.copy(json_obj)
            kwargs['model_name'] = json_obj['model_name']
            kwargs['serving_dir'] = json_obj['serving_dir']
            kwargs['conf_threshold'] = float(json_obj['conf_threshold'])
        except ValueError as e:
            raise ValueError(
                'Failed to convert json object to {} instance. '
                'The input json object is {}. ({})'.format(cls.__name__,
                                                           json_obj, e))
        return cls(**kwargs)

    def __call__(self, image):
        if not self.predictor:
            self.predictor = tfutils.TFServingPredictor('localhost', self.container_external_port)
        rgb_image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        results = self.predictor.infer_one(self.model_name, rgb_image, conf_threshold=self.conf_threshold)

        # # debug
        # debug_image = visualize_detections(image, results)
        # cv2.imshow('debug', debug_image)
        # cv2.waitKey(1)
        return results

    def clean(self):
        self.container_manager.clean()


if __name__ == "__main__":
    container_image_url = 'registry.cmusatyalab.org/junjuew/container-registry:tpod-image-sandwich-sandwich'
    processor = FasterRCNNContainerCallable(container_image_url=container_image_url,
                                            conf_threshold=0.1)
    image = cv2.imread('test.png')
    print(processor(image))
    processor.clean()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: wca-state-machine.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='wca-state-machine.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x17wca-state-machine.proto\"\xcb\x01\n\x13TransitionPredicate\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\rcallable_name\x18\x02 \x01(\t\x12\x41\n\x0f\x63\x61llable_kwargs\x18\x03 \x03(\x0b\x32(.TransitionPredicate.CallableKwargsEntry\x12\x15\n\rcallable_args\x18\x04 \x01(\t\x1a\x35\n\x13\x43\x61llableKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\"H\n\x0bInstruction\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x61udio\x18\x02 \x01(\t\x12\r\n\x05image\x18\x03 \x01(\x0c\x12\r\n\x05video\x18\x04 \x01(\x0c\"{\n\nTransition\x12\x0c\n\x04name\x18\x01 \x01(\t\x12(\n\npredicates\x18\x02 \x03(\x0b\x32\x14.TransitionPredicate\x12!\n\x0binstruction\x18\x03 \x01(\x0b\x32\x0c.Instruction\x12\x12\n\nnext_state\x18\x04 \x01(\t\"\xb7\x01\n\tProcessor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\rcallable_name\x18\x02 \x01(\t\x12\x37\n\x0f\x63\x61llable_kwargs\x18\x03 \x03(\x0b\x32\x1e.Processor.CallableKwargsEntry\x12\x15\n\rcallable_args\x18\x04 \x01(\t\x1a\x35\n\x13\x43\x61llableKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\"W\n\x05State\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1e\n\nprocessors\x18\x02 \x03(\x0b\x32\n.Processor\x12 \n\x0btransitions\x18\x03 \x03(\x0b\x32\x0b.Transition\"\xa3\x01\n\x0cStateMachine\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x16\n\x06states\x18\x02 \x03(\x0b\x32\x06.State\x12)\n\x06\x61ssets\x18\x03 \x03(\x0b\x32\x19.StateMachine.AssetsEntry\x12\x13\n\x0bstart_state\x18\x04 \x01(\t\x1a-\n\x0b\x41ssetsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\x62\x06proto3')
)




_TRANSITIONPREDICATE_CALLABLEKWARGSENTRY = _descriptor.Descriptor(
  name='CallableKwargsEntry',
  full_name='TransitionPredicate.CallableKwargsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='TransitionPredicate.CallableKwargsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='TransitionPredicate.CallableKwargsEntry.value', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=178,
  serialized_end=231,
)

_TRANSITIONPREDICATE = _descriptor.Descriptor(
  name='TransitionPredicate',
  full_name='TransitionPredicate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='TransitionPredicate.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='callable_name', full_name='TransitionPredicate.callable_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='callable_kwargs', full_name='TransitionPredicate.callable_kwargs', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='callable_args', full_name='TransitionPredicate.callable_args', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_TRANSITIONPREDICATE_CALLABLEKWARGSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28,
  serialized_end=231,
)


_INSTRUCTION = _descriptor.Descriptor(
  name='Instruction',
  full_name='Instruction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Instruction.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='audio', full_name='Instruction.audio', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='image', full_name='Instruction.image', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='video', full_name='Instruction.video', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=233,
  serialized_end=305,
)


_TRANSITION = _descriptor.Descriptor(
  name='Transition',
  full_name='Transition',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Transition.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='predicates', full_name='Transition.predicates', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='instruction', full_name='Transition.instruction', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='next_state', full_name='Transition.next_state', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=307,
  serialized_end=430,
)


_PROCESSOR_CALLABLEKWARGSENTRY = _descriptor.Descriptor(
  name='CallableKwargsEntry',
  full_name='Processor.CallableKwargsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Processor.CallableKwargsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='Processor.CallableKwargsEntry.value', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=178,
  serialized_end=231,
)

_PROCESSOR = _descriptor.Descriptor(
  name='Processor',
  full_name='Processor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Processor.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='callable_name', full_name='Processor.callable_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='callable_kwargs', full_name='Processor.callable_kwargs', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='callable_args', full_name='Processor.callable_args', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_PROCESSOR_CALLABLEKWARGSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=433,
  serialized_end=616,
)


_STATE = _descriptor.Descriptor(
  name='State',
  full_name='State',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='State.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='processors', full_name='State.processors', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transitions', full_name='State.transitions', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=618,
  serialized_end=705,
)


_STATEMACHINE_ASSETSENTRY = _descriptor.Descriptor(
  name='AssetsEntry',
  full_name='StateMachine.AssetsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='StateMachine.AssetsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='StateMachine.AssetsEntry.value', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=826,
  serialized_end=871,
)

_STATEMACHINE = _descriptor.Descriptor(
  name='StateMachine',
  full_name='StateMachine',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='StateMachine.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='states', full_name='StateMachine.states', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='assets', full_name='StateMachine.assets', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start_state', full_name='StateMachine.start_state', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_STATEMACHINE_ASSETSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=708,
  serialized_end=871,
)

_TRANSITIONPREDICATE_CALLABLEKWARGSENTRY.containing_type = _TRANSITIONPREDICATE
_TRANSITIONPREDICATE.fields_by_name['callable_kwargs'].message_type = _TRANSITIONPREDICATE_CALLABLEKWARGSENTRY
_TRANSITION.fields_by_name['predicates'].message_type = _TRANSITIONPREDICATE
_TRANSITION.fields_by_name['instruction'].message_type = _INSTRUCTION
_PROCESSOR_CALLABLEKWARGSENTRY.containing_type = _PROCESSOR
_PROCESSOR.fields_by_name['callable_kwargs'].message_type = _PROCESSOR_CALLABLEKWARGSENTRY
_STATE.fields_by_name['processors'].message_type = _PROCESSOR
_STATE.fields_by_name['transitions'].message_type = _TRANSITION
_STATEMACHINE_ASSETSENTRY.containing_type = _STATEMACHINE
_STATEMACHINE.fields_by_name['states'].message_type = _STATE
_STATEMACHINE.fields_by_name['assets'].message_type = _STATEMACHINE_ASSETSENTRY
DESCRIPTOR.message_types_by_name['TransitionPredicate'] = _TRANSITIONPREDICATE
DESCRIPTOR.message_types_by_name['Instruction'] = _INSTRUCTION
DESCRIPTOR.message_types_by_name['Transition'] = _TRANSITION
DESCRIPTOR.message_types_by_name['Processor'] = _PROCESSOR
DESCRIPTOR.message_types_by_name['State'] = _STATE
DESCRIPTOR.message_types_by_name['StateMachine'] = _STATEMACHINE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TransitionPredicate = _reflection.GeneratedProtocolMessageType('TransitionPredicate', (_message.Message,), dict(

  CallableKwargsEntry = _reflection.GeneratedProtocolMessageType('CallableKwargsEntry', (_message.Message,), dict(
    DESCRIPTOR = _TRANSITIONPREDICATE_CALLABLEKWARGSENTRY,
    __module__ = 'wca_state_machine_pb2'
    # @@protoc_insertion_point(class_scope:TransitionPredicate.CallableKwargsEntry)
    ))
  ,
  DESCRIPTOR = _TRANSITIONPREDICATE,
  __module__ = 'wca_state_machine_pb2'
  # @@protoc_insertion_point(class_scope:TransitionPredicate)
  ))
_sym_db.RegisterMessage(TransitionPredicate)
_sym_db.RegisterMessage(TransitionPredicate.CallableKwargsEntry)

Instruction = _reflection.GeneratedProtocolMessageType('Instruction', (_message.Message,), dict(
  DESCRIPTOR = _INSTRUCTION,
  __module__ = 'wca_state_machine_pb2'
  # @@protoc_insertion_point(class_scope:Instruction)
  ))
_sym_db.RegisterMessage(Instruction)

Transition = _reflection.GeneratedProtocolMessageType('Transition', (_message.Message,), dict(
  DESCRIPTOR = _TRANSITION,
  __module__ = 'wca_state_machine_pb2'
  # @@protoc_insertion_point(class_scope:Transition)
  ))
_sym_db.RegisterMessage(Transition)

Processor = _reflection.GeneratedProtocolMessageType('Processor', (_message.Message,), dict(

  CallableKwargsEntry = _reflection.GeneratedProtocolMessageType('CallableKwargsEntry', (_message.Message,), dict(
    DESCRIPTOR = _PROCESSOR_CALLABLEKWARGSENTRY,
    __module__ = 'wca_state_machine_pb2'
    # @@protoc_insertion_point(class_scope:Processor.CallableKwargsEntry)
    ))
  ,
  DESCRIPTOR = _PROCESSOR,
  __module__ = 'wca_state_machine_pb2'
  # @@protoc_insertion_point(class_scope:Processor)
  ))
_sym_db.RegisterMessage(Processor)
_sym_db.RegisterMessage(Processor.CallableKwargsEntry)

State = _reflection.GeneratedProtocolMessageType('State', (_message.Message,), dict(
  DESCRIPTOR = _STATE,
  __module__ = 'wca_state_machine_pb2'
  # @@protoc_insertion_point(class_scope:State)
  ))
_sym_db.RegisterMessage(State)

StateMachine = _reflection.GeneratedProtocolMessageType('StateMachine', (_message.Message,), dict(

  AssetsEntry = _reflection.GeneratedProtocolMessageType('AssetsEntry', (_message.Message,), dict(
    DESCRIPTOR = _STATEMACHINE_ASSETSENTRY,
    __module__ = 'wca_state_machine_pb2'
    # @@protoc_insertion_point(class_scope:StateMachine.AssetsEntry)
    ))
  ,
  DESCRIPTOR = _STATEMACHINE,
  __module__ = 'wca_state_machine_pb2'
  # @@protoc_insertion_point(class_scope:StateMachine)
  ))
_sym_db.RegisterMessage(StateMachine)
_sym_db.RegisterMessage(StateMachine.AssetsEntry)


_TRANSITIONPREDICATE_CALLABLEKWARGSENTRY._options = None
_PROCESSOR_CALLABLEKWARGSENTRY._options = None
_STATEMACHINE_ASSETSENTRY._options = None
# @@protoc_insertion_point(module_scope)



"""Base class and helper functions for callable classes.
"""
import inspect
from functools import wraps


def record_kwargs(func):
    """
    Decorator to automatically record constructor arguments

    >>> class process:
    ...     @record_kwargs
    ...     def __init__(self, cmd, reachable=False, user='root'):
    ...         pass
    >>> p = process('halt', True)
    >>> p.cmd, p.reachable, p.user
    ('halt', True, 'root')
    """
    names, varargs, keywords, defaults = inspect.getargspec(func)
    if defaults is None:
        defaults = ()

    @wraps(func)
    def wrapper(self, *args, **kargs):
        func(self, *args, **kargs)
        kwargs = {}
        for name, default in zip(reversed(names), reversed(defaults)):
            kwargs[name] = default
        for name, arg in list(zip(names[1:], args)) + list(kargs.items()):
            kwargs[name] = arg
        setattr(self, 'kwargs', kwargs)

    return wrapper


class CallableBase():
    """Base class for Callables used in FSMs.

    Custom callable classes need to inherit from this class. Inherited classes
    should add decorator @record_kwargs to their constructors for serialization.
    """

    def __init__(self):
        super().__init__()
        setattr(self, 'kwargs', {})

    @classmethod
    def from_json(cls, json_obj):
        """Create a CallableBase class instance from a json object.

        Subclasses should overide this class depending on the input type of
        their constructor.
        """
        return cls(**json_obj)

    def __eq__(self, other):
        if isinstance(other, CallableBase):
            return self.kwargs == other.kwargs
        return False

    def __ne__(self, other):
        return not self.__eq__(other)


class Null(CallableBase):
    """A empty callable class that returns None.

    Useful for initialization.
    """

    def __call__(self, *args, **kwargs):
        return None



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: instruction.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='instruction.proto',
  package='instruction',
  syntax='proto3',
  serialized_pb=_b('\n\x11instruction.proto\x12\x0binstruction\"\xf6\x01\n\x04Ikea\x12&\n\x05state\x18\x01 \x01(\x0e\x32\x17.instruction.Ikea.State\x12\x1e\n\x16\x66rames_with_one_buckle\x18\x02 \x01(\x05\x12\x1f\n\x17\x66rames_with_two_buckles\x18\x03 \x01(\x05\"\x84\x01\n\x05State\x12\t\n\x05START\x10\x00\x12\x0b\n\x07NOTHING\x10\x01\x12\x08\n\x04\x42\x41SE\x10\x02\x12\x08\n\x04PIPE\x10\x03\x12\t\n\x05SHADE\x10\x04\x12\n\n\x06\x42UCKLE\x10\x05\x12\x10\n\x0c\x42LACK_CIRCLE\x10\x06\x12\x0e\n\nSHADE_BASE\x10\x07\x12\x08\n\x04\x42ULB\x10\x08\x12\x0c\n\x08\x42ULB_TOP\x10\t\"\xe9\x01\n\x08Sandwich\x12*\n\x05state\x18\x01 \x01(\x0e\x32\x1b.instruction.Sandwich.State\x12\x0e\n\x06holo_x\x18\x02 \x01(\x01\x12\x0e\n\x06holo_y\x18\x03 \x01(\x01\x12\x12\n\nholo_depth\x18\x04 \x01(\x01\"}\n\x05State\x12\t\n\x05START\x10\x00\x12\x0b\n\x07NOTHING\x10\x01\x12\t\n\x05\x42READ\x10\x02\x12\x07\n\x03HAM\x10\x03\x12\x0b\n\x07LETTUCE\x10\x04\x12\x0c\n\x08\x43UCUMBER\x10\x05\x12\x08\n\x04HALF\x10\x06\x12\n\n\x06TOMATO\x10\x07\x12\r\n\tHAM_WRONG\x10\x08\x12\x08\n\x04\x46ULL\x10\t\"{\n\x0c\x45ngineFields\x12\x14\n\x0cupdate_count\x18\x01 \x01(\x03\x12!\n\x04ikea\x18\x02 \x01(\x0b\x32\x11.instruction.IkeaH\x00\x12)\n\x08sandwich\x18\x03 \x01(\x0b\x32\x15.instruction.SandwichH\x00\x42\x07\n\x05stateB(\n\x1e\x65\x64u.cmu.cs.gabriel.instructionB\x06Protosb\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)



_IKEA_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='instruction.Ikea.State',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='START', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NOTHING', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BASE', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PIPE', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SHADE', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BUCKLE', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BLACK_CIRCLE', index=6, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SHADE_BASE', index=7, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BULB', index=8, number=8,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BULB_TOP', index=9, number=9,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=149,
  serialized_end=281,
)
_sym_db.RegisterEnumDescriptor(_IKEA_STATE)

_SANDWICH_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='instruction.Sandwich.State',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='START', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NOTHING', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BREAD', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HAM', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LETTUCE', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUCUMBER', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HALF', index=6, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TOMATO', index=7, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HAM_WRONG', index=8, number=8,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FULL', index=9, number=9,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=392,
  serialized_end=517,
)
_sym_db.RegisterEnumDescriptor(_SANDWICH_STATE)


_IKEA = _descriptor.Descriptor(
  name='Ikea',
  full_name='instruction.Ikea',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='state', full_name='instruction.Ikea.state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='frames_with_one_buckle', full_name='instruction.Ikea.frames_with_one_buckle', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='frames_with_two_buckles', full_name='instruction.Ikea.frames_with_two_buckles', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _IKEA_STATE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=35,
  serialized_end=281,
)


_SANDWICH = _descriptor.Descriptor(
  name='Sandwich',
  full_name='instruction.Sandwich',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='state', full_name='instruction.Sandwich.state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='holo_x', full_name='instruction.Sandwich.holo_x', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='holo_y', full_name='instruction.Sandwich.holo_y', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='holo_depth', full_name='instruction.Sandwich.holo_depth', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SANDWICH_STATE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=284,
  serialized_end=517,
)


_ENGINEFIELDS = _descriptor.Descriptor(
  name='EngineFields',
  full_name='instruction.EngineFields',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='update_count', full_name='instruction.EngineFields.update_count', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ikea', full_name='instruction.EngineFields.ikea', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sandwich', full_name='instruction.EngineFields.sandwich', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='state', full_name='instruction.EngineFields.state',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=519,
  serialized_end=642,
)

_IKEA.fields_by_name['state'].enum_type = _IKEA_STATE
_IKEA_STATE.containing_type = _IKEA
_SANDWICH.fields_by_name['state'].enum_type = _SANDWICH_STATE
_SANDWICH_STATE.containing_type = _SANDWICH
_ENGINEFIELDS.fields_by_name['ikea'].message_type = _IKEA
_ENGINEFIELDS.fields_by_name['sandwich'].message_type = _SANDWICH
_ENGINEFIELDS.oneofs_by_name['state'].fields.append(
  _ENGINEFIELDS.fields_by_name['ikea'])
_ENGINEFIELDS.fields_by_name['ikea'].containing_oneof = _ENGINEFIELDS.oneofs_by_name['state']
_ENGINEFIELDS.oneofs_by_name['state'].fields.append(
  _ENGINEFIELDS.fields_by_name['sandwich'])
_ENGINEFIELDS.fields_by_name['sandwich'].containing_oneof = _ENGINEFIELDS.oneofs_by_name['state']
DESCRIPTOR.message_types_by_name['Ikea'] = _IKEA
DESCRIPTOR.message_types_by_name['Sandwich'] = _SANDWICH
DESCRIPTOR.message_types_by_name['EngineFields'] = _ENGINEFIELDS

Ikea = _reflection.GeneratedProtocolMessageType('Ikea', (_message.Message,), dict(
  DESCRIPTOR = _IKEA,
  __module__ = 'instruction_pb2'
  # @@protoc_insertion_point(class_scope:instruction.Ikea)
  ))
_sym_db.RegisterMessage(Ikea)

Sandwich = _reflection.GeneratedProtocolMessageType('Sandwich', (_message.Message,), dict(
  DESCRIPTOR = _SANDWICH,
  __module__ = 'instruction_pb2'
  # @@protoc_insertion_point(class_scope:instruction.Sandwich)
  ))
_sym_db.RegisterMessage(Sandwich)

EngineFields = _reflection.GeneratedProtocolMessageType('EngineFields', (_message.Message,), dict(
  DESCRIPTOR = _ENGINEFIELDS,
  __module__ = 'instruction_pb2'
  # @@protoc_insertion_point(class_scope:instruction.EngineFields)
  ))
_sym_db.RegisterMessage(EngineFields)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\036edu.cmu.cs.gabriel.instructionB\006Protos'))
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
"""Callable classes for Transition Predicates.

All the classes here should be a callable and return either True/False when
called (to indicate whether or not to take a transition). All classes should
inherit from CallableBase class and annoate their constructor (if there is one)
with the @record_kwargs decorator for proper serialization.
"""
import time

from gabrieltool.statemachine.callable_zoo import record_kwargs
from gabrieltool.statemachine.callable_zoo import CallableBase


class HasObjectClass(CallableBase):
    """Check if there is an object class in the extracted information of the sensor data.
    """

    @record_kwargs
    def __init__(self, class_name):
        """Constructor.

        Args:
            class_name (string): Class name or id.
        """
        super().__init__()
        self.class_name = class_name

    def __call__(self, app_state):
        return self.class_name in app_state


class Always(CallableBase):
    """Always take this transition.

    Useful for welcome message when the application starts.
    """

    def __call__(self, app_state):
        return True


class HasObjectClassWhileNotOthers(CallableBase):
    """Check if there are some object classes in the extracted information while some other classes are not.
    """

    @record_kwargs
    def __init__(self, has_classes=None, absent_classes=None):
        """Constructor.

        Args:
            existed_classes (list of strings, optional): Names of classes that
                need to exist. Defaults to None.
            absent_classes (list of strings, optional): Name of classes that
                need to be absent. Defaults to None.
        """
        super().__init__()
        self.has_classes = has_classes if has_classes is not None else []
        self.absent_classes = absent_classes if absent_classes is not None else []

    def __call__(self, app_state):
        return all(
            [class_name in app_state for class_name in self.has_classes]) and all(
                [class_name not in app_state for class_name in self.absent_classes]
        )


class Wait(CallableBase):
    """Wait for some time before turning true.
    """

    @record_kwargs
    def __init__(self, wait_time=None):
        """Constructor.

        Args:
            wait_time (int, optional): Wait time in seconds. Defaults to None.
        """
        super().__init__()
        self.wait_time = wait_time if wait_time is not None else 0
        # set when this predicate this first called
        # or returned True in the last call.
        self._start_time = None

    def __call__(self, app_state):
        if self._start_time is None:
            self._start_time = time.time()
        else:
            cur_time = time.time()
            if (cur_time - self._start_time) > self.wait_time:
                # reset
                self._start_time = None
                return True
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'pip>=20',
    'setuptools>=46',
    'tensorflow-serving-api~=1.15',
    'docker',
    'opencv-python>=3',
    'logzero',
    'fire',
    'gabriel-server',
]

setup_requirements = []

test_requirements = ['pytest', ]

setup(
    author="Junjue Wang",
    author_email='junjuew@cs.cmu.edu',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    python_requires=">=3.5, <3.8",
    description="Tools for Making Wearable Cognitive Assitants",
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='gabrieltool',
    name='gabrieltool',
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/junjuew/gabriel-tool',
    version='1.0.5',
    zip_safe=False,
)



# -*- coding: utf-8 -*-
"""Finite State Machine Runner.

Runner to run the cognitive assistants that are expressed as state machines.
"""

import cv2
import numpy as np
from gabriel_protocol import gabriel_pb2
from gabriel_server import cognitive_engine
from logzero import logger

from gabrieltool.statemachine import fsm, instruction_pb2


class Runner(object):
    """Finite State Machine Runner.

    A basic finite state machine runner.
    Make sure the fsm is constructed fully before creating a runner.
    """

    def __init__(self, start_state, prepare_to_run=True):
        """Construct a FSM runner.

        Args:
            start_state (State): The start state of a FSM.
            prepare_to_run (bool, optional): Whether to call prepare() functions
                on all state before running. It should be set to true unless debugging.
                Defaults to True.
        """
        super(Runner, self).__init__()
        self.current_state = start_state
        if prepare_to_run:
            self._prepare_to_run()

    def feed(self, data, debug=False):
        """Feed the FSM an input to get an output.

        Args:
            data (any): Input data.
            debug (bool, optional): Defaults to False.

        Raises:
            ValueError: when current state is None.

        Returns:
            Instruction: Instruction from the FSM.
        """
        if self.current_state is None:
            raise ValueError('Current State is None! Did you forget to specify transition\'s next_state?')
        next_state, instruction = self.current_state(data)
        self.current_state = next_state
        return instruction

    def _prepare_to_run(self):
        """Prepare each state in the state machine to run.

        This allows each state to load asset from disks, start container, etc.
        Each state's behavior should be implemented in their prepare function.
        """
        for state in fsm.StateMachine.bfs(self.current_state):
            state.prepare()


class BasicCognitiveEngineRunner(cognitive_engine.Engine):
    """A basic Gabriel Cognitive Engine Runner for FSM based cognitive assistants.

    This runner will start a Gabriel cognitive engine that follows the logic
    defined in a FSM. This class is basic as it restricts the sensor input to be
    images and the instruction output to be audio or images.
    """

    def __init__(self, engine_name, fsm):
        """Construct a Gabriel Cognitive Engine Runner.

        Args:
            engine_name (string): Name of the cognitive engine.
            fsm (State): The start state of an FSM.
        """
        super(BasicCognitiveEngineRunner, self).__init__()
        self.engine_name = engine_name
        self._fsm = fsm
        self._fsm_runner = Runner(self._fsm)

    def handle(self, from_client):
        """Do not call directly.

        This method is invoked by the gabriel framework when a user input is available.
        """
        if from_client.payload_type != gabriel_pb2.PayloadType.Value('IMAGE'):
            return cognitive_engine.wrong_input_format_error(
                from_client.frame_id)
        engine_fields = cognitive_engine.unpack_engine_fields(
            instruction_pb2.EngineFields, from_client)

        img_array = np.asarray(bytearray(from_client.payload), dtype=np.int8)
        img = cv2.imdecode(img_array, -1)

        inst = self._fsm_runner.feed(img)

        result_wrapper = gabriel_pb2.ResultWrapper()
        engine_fields.update_count += 1
        result_wrapper.engine_fields.Pack(engine_fields)

        if inst.image:
            result = result_wrapper.results.add()
            result.payload_type = gabriel_pb2.PayloadType.Value('IMAGE')
            result.payload = inst.image
            result.engine_name = self.engine_name

        if inst.audio:
            result = result_wrapper.results.add()
            result.payload_type = gabriel_pb2.PayloadType.Value('TEXT')
            result.payload = inst.audio.encode(encoding="utf-8")
            result.engine_name = self.engine_name

        logger.info('Current State: {}'.format(self._fsm_runner.current_state))
        logger.info(result_wrapper.results)

        result_wrapper.frame_id = from_client.frame_id
        result_wrapper.status = gabriel_pb2.ResultWrapper.Status.Value('SUCCESS')

        return result_wrapper
